from huggingface_hub import (
    hf_hub_download,
    list_repo_files
)
import os
from pathlib import Path
import polars as pl
import random
from shutil import rmtree
from tld import get_tld
from tqdm.auto import tqdm
from typing import List, Dict
from urllib.parse import urlparse

def get_fineweb_paths(language = "")-> List[str]:
    """Helper function that queries the FineWeb repository and returns all train files."""
    paths =  [
        path
        for path in list_repo_files("HuggingFaceFW/fineweb-2", repo_type="dataset")
        if path.endswith("parquet")
        and "removed" not in path
        and "train" in path
    ]
    if language:
        paths = [path for path in paths if language in path]
    return paths

def get_fineweb_languages()-> List[str]:
    """Helper function that collects all languages from the FineWeb repository."""
    paths = get_fineweb_paths()
    return list(set([path.split('/')[1] for path in paths]))

def get_io_paths(language, output_dir: str='FineWeb')-> Dict[str, Path]:
    """Helper function for building directory paths to save raw, intermediate, and preprocessed data."""

    dirs = {
        'stripped': Path(output_dir, 'intermediate', language, 'stripped'),
        'grouped': Path(output_dir, 'intermediate', language, 'grouped'),
        'preprocessed': Path(output_dir, 'preprocessed', f'{language}.parquet')
    }
    return dirs

def download_fineweb_data(
    language: str,
    local_dir: str = 'FineWeb/raw',
    sample: int = int(),
    disable_progress_bar: bool = False,
    restart: bool = False
)-> None:
    
    """
    Given a language code, this function will download raw FineWeb files from HF Hub.
    For possible languages, please use 'get_fineweb_languages', or visit https://huggingface.co/datasets/HuggingFaceFW/fineweb-2.
    Use the 'sample' arguement to run a sanity check.

    Args:
        language (str): Language code of the target language.
        local_dir (str): Directory where raw FineWeb data is to be saved.
        sample (int): For running a sanity check. Will sample 'n' paths.
        diable_progress_bar (bool): Whether to diable the progress bar.
        restart (bool): If True, will download all paths and overwrite previously written files.
    """
    
    #Checks if input language code matches FineWeb data.
    print(f'Checking FineWeb hub for the language code, "{language}"...')
    possible_languages = get_fineweb_languages()
    if language not in possible_languages:
        raise ValueError ("Input language is not recognized. Please use the function, 'get_fine_web_languages' to view possible languages, or visit https://huggingface.co/datasets/HuggingFaceFW/fineweb-2")

    #Retrieves paths. If doing a sanity check, takes a sample of the paths for downloading.

    paths = get_fineweb_paths(language)
    if sample:
        sample = min(sample, len(paths))
        paths = random.sample(paths, sample)
        print (f'Running sanity check. {sample} paths sampled.')
    os.environ["HF_HUB_ENABLE_HF_TRANSFER"] = "1"
    
    #Creates a directory to save FineWeb data.
    os.makedirs(local_dir, exist_ok=True)

    #Iterates through paths and downloads FineWeb data using 'hf_hub_download'.
    for path in tqdm(paths, desc="Downloading FineWeb data", disable=disable_progress_bar):
        if os.path.exists(Path(local_dir, path)) and not restart:
            continue

        hf_hub_download(
        repo_id="HuggingFaceFW/fineweb-2",
        repo_type="dataset",
        filename=path,
        local_dir=local_dir)
    
    raw_data_dir = Path(local_dir, os.path.dirname(paths[0]))
    print(f'Download complete. Saved raw FineWeb data to {raw_data_dir}')
    return raw_data_dir

def get_domain(url: str)-> str:
    """Helper function that will take a URL and return the domain."""

    parsed_url = urlparse(url)
    return f"{parsed_url.scheme}://{parsed_url.netloc}"

def add_domain_column(df: pl.DataFrame)-> pl.DataFrame:
    """Helper function that will take a polars FineWeb DataFrame and add a domain column."""

    df = df.with_columns(
    pl.col("url").map_elements(lambda x: get_domain(x), return_dtype=pl.Utf8).alias("domain")
    )
    return df.drop_nans()

def strip_and_add_domain_pipeline(
        language: str,
        raw_data_dir: str,
        output_dir: str = 'FineWeb',
        columns: List[str] = ['id', 'date', 'url'],
        disable_progress_bar: bool = False,
        print_statement: bool = True,
        restart: bool = False
)-> None:
    """
    Pipeline that will iterate through FineWeb parquet files,
    strip away unneccesary data, add the domain column, and save
    the DataFrame to disk.

    Args:
        language (str): Language which is being processed.
        raw_data_dr (str): Explicit path to directory where raw data is stored.
        output_dir (str): Direcotry to save intermediate and final outputs.
        columns (List[str]): The columns from FineWeb that are to be kept.
        disable_progress_bar (bool): Whether or not to disable the progresss bar.
        print_statement(bool): Whether to issue a print statment when complete.
        restart (bool): If True, starts over and overwrites files that have already been processed.
    """

    #Build your input/output paths
    io_paths = get_io_paths(language, output_dir)

    #Checks if FineWeb data has been stripped with domain column added.
    if not os.path.exists(raw_data_dir):
        raise FileNotFoundError(f'{raw_data_dir} not found. Use download_fineweb_data to get language data.')
    elif len(os.listdir(raw_data_dir)) == 0:
        raise FileNotFoundError(f'{raw_data_dir} found, but it has no files. Use download_fineweb_data to get language data.')
 
    #Builds a list a files, and makes a directory to save stripped data.
    files = [file for file in os.listdir(raw_data_dir) if file.endswith('.parquet')]
    os.makedirs(io_paths['stripped'], exist_ok=True)

    #Ensures that 'url' is included, which is neccesary to proceed with grouping.
    if 'url' not in columns:
        raise ValueError ("To proceed with grouping, 'url' must be included in columns.")
    
    #Ensures that the input columns are contained in the FineWeb data.
    possible_columns = pl.read_parquet_schema(Path(raw_data_dir, os.listdir(raw_data_dir)[0]))
    if not all(column in possible_columns for column in columns):
        raise ValueError (f"Column selection does not align with FineWeb data. Please select from these columns: {list(possible_columns.keys())}")

    for file in tqdm(files, desc= "Stripping dfs and adding domain column", disable=disable_progress_bar):
        input_file, output_file = Path(raw_data_dir, file), Path(io_paths["stripped"], file)
        
        if os.path.exists(output_file) and not restart:
            continue

        #Iterates through files, loads only the specefied columns, adds the domain column, and saves to disk.
        df = pl.read_parquet(input_file, columns=columns)
        df = add_domain_column(df)
        df.write_parquet(output_file)
    if print_statement:
        total_files = len(os.listdir(io_paths['stripped']))
        print(f'Stripping complete. {total_files} processed. Saved files to {io_paths["stripped"]}')

def group_pipeline(
    language: str,
    output_dir: str="FineWeb",
    disable_progress_bar: bool = False,
    print_statement: bool = True,
    restart: bool = False
)-> None:
    """
    Given a directory of parquet files that have been processed by the function,
    'strip_and_add_domain_pipeline', this function will group them by domain
    and save them to disk.

    Args:
        language (str): Language which is being processed.
        output_dir (str): Direcotry to save intermediate and final outputs.
        disable_progress_bar (bool): Whether or not to disable the progress bar.
        print_statment (bool): Whether or not to issue a print statemnt on completion.
        restart (bool): If True, starts over and overwrites files that have already been processed.

    """
    #Build your input/output paths
    io_paths = get_io_paths(language, output_dir)

    if not os.path.exists(io_paths['stripped']):
        raise FileNotFoundError(f'{io_paths["stripped"]} not found. Run strip_and_add_domain_pipeline before grouping.')
    elif len(os.listdir(io_paths['stripped'])) == 0:
        raise FileNotFoundError(f'{io_paths["stripped"]} found, but it has no files. Run strip_and_add_domain_pipeline before grouping.')

    #Builds a list a files, and makes a directory to save stripped data.
    files = [file for file in os.listdir(io_paths['stripped']) if file.endswith('.parquet')]
    os.makedirs(io_paths['grouped'], exist_ok=True)

    for file in tqdm(files, desc= 'Grouping by domain', disable=disable_progress_bar):
        input_file, output_file = Path(io_paths['stripped'], file), Path(io_paths['grouped'], file)
        if os.path.exists(output_file) and not restart: 
                continue
        
        #Loads the file, groups by domain, returning a df with columns 'domain' and 'count'
        df = pl.read_parquet(input_file)
        df = df.group_by('domain').count()
        df.write_parquet(output_file)
    if print_statement:
        total_files = len(os.listdir(io_paths['grouped']))
        print(f'Grouping complete. {total_files} processed. Saved output to {io_paths["grouped"]}')

def combine_grouped_dfs(
    language: str,
    output_dir: str='FineWeb',
    batch_size: int=10,
    disable_progress_bar: bool = False,
    print_statement: bool = True
    )-> None:
    """
    Given a directory of grouped DataFrames, combines into a single DataFrame.

    Args:
        language (str): Language which is being processed.
        output_dir (str): Direcotry to save intermediate and final outputs.
        batch_size (int): Number of DataFrames to process in each batch.
        diable_progress_bar (bool): Whether to disable the progress bar.
        print_statement (bool): Whether to issue a print statement on completion.
    
    Returns (saved to disk):
        Grouped DataFrame with the following columns:
            domain (str): URL of the contributing domain.
            count (u32): Number of pages which that domain contributed.
            tld (str): Top-level domain of the contributing domain.
    """

    #Build your input/output paths
    io_paths = get_io_paths(language, output_dir)

    #Checks if FineWeb data has been stripped with domain column added.
    if not os.path.exists(io_paths['grouped']):
        raise FileNotFoundError(f'{io_paths["grouped"]} not found. Group data using group_pipeline.')
    elif len(os.listdir(io_paths['grouped'])) == 0:
        raise FileNotFoundError(f'{io_paths["grouped"]} found, but it has no files. Group data using group_pipeline.')
    #Initialize the output and the filepaths
    output = None 
    paths = [Path(io_paths['grouped'], file) for file in os.listdir(io_paths['grouped']) if file.endswith('.parquet')]

    #Compile the data in batches
    for i in tqdm(
        range(0, len(paths), batch_size),
        desc = "Combining grouped data",
        disable=disable_progress_bar):

        #Concatenate a batch of files and group them by domain.
        batch_files = paths[i:i + batch_size]
        batch_data = pl.concat([pl.read_parquet(file) for file in batch_files])
        batch_grouped = batch_data.group_by('domain').agg(pl.col('count').sum())
        
        #Combine current batch with previous batches.
        if output is None:
            output = batch_grouped
        else:
            output = pl.concat([output, batch_grouped]).group_by('domain').agg(pl.col('count').sum())
    
    if print_statement:
        print ("All DataFrames combined. Adding top-level domain column.")
    
    #Add the tld column.
    output = output.with_columns(
        pl.col("domain").map_elements(lambda x: get_tld(x, fail_silently=True), return_dtype=pl.Utf8).alias("tld")
        )
    
    #Sort and save
    os.makedirs(os.path.dirname(io_paths['preprocessed']), exist_ok=True)
    output = output.sort('count', descending=True)
    output.write_parquet(io_paths['preprocessed'])

    if print_statement:
        print(f"Saved combined dataframe to {io_paths['preprocessed']}")
        print(output.head(10))

def download_and_preprocess_pipeline(
    language: str,
    local_dir: str = "FineWeb/raw",
    output_dir: str = 'FineWeb',
    sample: int = int(),
    columns: List[str] = ['id', 'date', 'url'],
    batch_size: int=10,
    disable_progress_bar: bool = False,
    print_statement: bool = True,
    restart: bool = False
    )-> pl.DataFrame:
    """
    Combines all previous pipelines.
    Given a language, this function will download FineWeb data, strip it,
    group it, and combine all files into a single DataFrame.

    Args:
        language (str): Language which is being processed.
        local_dir (str): Directory where raw FineWeb data is stored.
        output_dir (str): Direcotry to save intermediate and final outputs.
        columns (List[str]): During stripping, identifies columns to keep.
        grouped_data_dir (str): Directory to save DataFrames grouped by domain.
        batch_size (int): When combining grouped data, the number of DataFrames to combine in each batch.
        path_to_output (str): Path to save the final combined DataFrame.
        disable_progress_bar (bool): Whether or not to disable the progress bar.
        print_statment (bool): Whether or not to issue a print statemnt on completion.
        restart (bool): If True, starts over and overwrites files that have already been processed.
    
    Returns:
        Each step of preprocessing saves files to disk. 
        With the default settings, the structure will be:

        root/
        └── FineWeb/
            ├── raw/
            │   └── data/
            │       └── {language}/
            │           └── train/
            │               ├── 000_00000.parquet
            │               ├── 000_00001.parquet
            │               └── 000_00002.parquet
            ├── intermediate/
            │   └── {language}/
            │       ├── stripped/
            │       │   ├── 000_00000.parquet
            │       │   ├── 000_00001.parquet
            │       │   └── 000_00002.parquet
            │       └── grouped/
            │           ├── 000_00000.parquet
            │           ├── 000_00001.parquet
            │           └── 000_00002.parquet
            └── preprocessed/
                └── {language}.parquet

        
        raw: Raw data from from FineWeb.
        intermediate/{lang}/stripped: DataFrames with unneccesary columns stripped and domain added.
        intermediate/{lang}/grouped: Invidual DataFrames grouped by domain.
        preprocessed/{lang}.parquet: Final output. All grouped DataFrames are combined and tld is added.
    """

    raw_data_dir = download_fineweb_data(
        language=language,
        local_dir=local_dir,
        sample=sample,
        disable_progress_bar=disable_progress_bar,
        restart=restart
    )

    strip_and_add_domain_pipeline(
        language=language,
        raw_data_dir=raw_data_dir,
        output_dir=output_dir,
        columns=columns,
        disable_progress_bar=disable_progress_bar,
        print_statement=print_statement,
        restart=restart
    )

    group_pipeline(
        language=language,
        output_dir=output_dir,
        disable_progress_bar=disable_progress_bar,
        print_statement=print_statement,
        restart=restart
    )

    combine_grouped_dfs(
        language=language,
        output_dir=output_dir,
        batch_size=batch_size,
        disable_progress_bar=disable_progress_bar,
        print_statement=print_statement
    )


import bs4
from collections import Counter
from datasets import Dataset
import json
import os
from pathlib import Path
import polars as pl
import re
import requests
from selenium import webdriver
from selenium.webdriver.common.by import By
from tqdm.auto import tqdm
from typing import Union, List, Literal, Tuple
from urllib.parse import urlparse

def filter_links(
    links: List[str],
    regex: str = ""
)->List[str]:
    """Helper funtion that will filter links using the provided regex."""
    
    if regex:
        links = [link for link in list(set(links)) if re.search(regex, link)]
    return links

def scrape_links(
    method: Literal['bs4', 'selenium'],
    url: str,
    regex: str = "",
    driver: object = None
) -> List[str]:
    """
    Scrapes links from a target page using either the BeautifulSoup library (bs4) or Selenium.

    Args:
        method (str): The scraping method to use ('bs4' or 'selenium').
        url (str): The URL from which to scrape links.
        regex (str, optional): If provided, regular expressions will be used to filter the links collected from the target page.
        driver (object, optional): A pre-initialized Selenium WebDriver instance. If not provided, a new instance will be created.

    Returns:
        links (list): A list of links scraped from the target page.
    """
    if method not in ['bs4', 'selenium']:
        raise ValueError('Input scraping method not supported. Please choose bs4 or selenium')
    
    if method == 'bs4':
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raise HTTPError for bad responses
            
            soup = bs4.BeautifulSoup(response.text, 'html.parser')

            # Find all <a> tags and extract their href attributes
            links = [a_tag['href'] for a_tag in soup.find_all('a', href=True)]
        
        except requests.exceptions.RequestException as e:
            print(f"Error fetching {url}: {e}")
            return []
        
    elif method == 'selenium':
        try:
            # If no driver is provided, create a new instance
            if not driver:
                driver = webdriver.Chrome()
            
            driver.get(url)

            # Locate all links on the page using the <a> tag
            links_elements = driver.find_elements(By.TAG_NAME, "a")

            # Extract href attributes from the <a> tags
            links = [link.get_attribute('href') for link in links_elements]
        
        except Exception as e:
            print(f"Error during Selenium scraping on {url}: {e}")
            links = []
        
        finally:
            # Ensure the driver quits after the scraping is done
            if driver:
                driver.quit()
    
    # Filter links with the provided regex
    return filter_links(links=links, regex=regex)

def count_path_branches(links: List[str])->Counter:
    """
    Useful for developing regex to target links.

    Takes a list of links and counts URL branches.
    """
    paths = [urlparse(link).path for link in links]
    branches = []
    
    # Split each path into hierarchical components
    for path in paths:
        parts = path.split('/')
        for i in range(1, len(parts) + 1):
            branches.append('/'.join(parts[:i]))
    
    # Count the frequency of each branch
    branch_freq = Counter([b for b in branches if b])
    return branch_freq

def save_list_to_txt(
        _list: List[str],
        output_path: str
) -> None:
    """Helper function that saves a list to a text file."""
    #checks if output path is to be saved in a directory, makes directory if neccesary.
    if os.path.exists(os.path.dirname(output_path)):
        os.makedirs(os.path.dirname(output_path), exist_ok=True)

    #As the crawl iterates across pages, scraped links are saved to a text file.
    if os.path.exists(output_path):
        with open(output_path, 'r', encoding ='utf-8') as f:

            #Combines new and old contents.
            old_list = f.read().split('\n')
            _list += old_list
    
    #Eliminates duplicates.
    _list = list(set(_list))

    #Writes the list to a text file, sepearted by newlines.
    with open(output_path, 'w', encoding = 'utf-8') as f:
        f.write('\n'.join(_list))

def load_list_from_txt(file_path: str)->List[str]:
    """Helper function that loads a text seperated by newlines into a list"""
    with open(file_path, 'r', encoding='utf-8') as f:
        ids = [id_ for id_ in f.read().split('\n') if id_ != ""]
        return ids
    
def pagerange(
        fstring: str, 
        first: int, 
        last: str
        )->List[str]:
    """
    Helper function for building a list of pages to crawl.
    
    Args:
        fstring (str): Base URL, with open curly-brackets placed where the page is input.
        first (int): The first page number.
        last (int): The last page number.

    Ex: pagerange('example.com/page{}, 1, 3) -> ['example.com/page1', 'example.com/page2', 'example.com/page3']
    """
    return [fstring.format(i) for i in range(first, last+1)]

def crawl(
        method: Literal['bs4', 'selenium'],
        urls : Union[str, List[str]],
        output_path: str,
        regex: str="",
        driver: object=None
) -> None:
    """
    Given a list of URLs, this functional will iterate across pages, scrape links, and save the list to a .txt file.

    Args:
        scrape_method (function): Identifies which library to use for scraping.
        urls (Union[str, List[str]]): Either a URL or a list of URLs to scrape.
        output_path (str): Path to save the list of links.
        regex (str): Regex used to filter URLs.
        driver (object): If using selenium, driver from which to access the web.
    """
    #Iterates across URLs, scrapes links, and saves them.
    if type(urls) == str:
        urls = [urls]
    for url in tqdm(urls, desc = 'Crawling and scraping links'):
        links = scrape_links(
            method=method, 
            url=url, 
            regex=regex,
            driver=driver
            )
        if links:
            save_list_to_txt(links, output_path)
    
    #Upon completion, states how many links were scraped.
    links = load_list_from_txt(output_path)
    print(f"Crawl complete. {len(links)} links scraped")

def extract_path(url: str) -> str:
    """Helper function that normalizes the URL for matching scraped links to FineWeb"""

    #Removes the terminal '/'
    if url[-1] == '/':
        url = url[:-1]
    if url[0] == '/':
        url = url[1:]

    # Check if the input is a full URL. If so, extracts the path.
    if re.match(r'^(https?:\/\/|www\.)', url):
        parsed_url = urlparse(url)
        return parsed_url.path if parsed_url.path else '/'
    
    # If not a full URL, simply returns the input.
    else:
        return url
    
def combine_links_into_regex(links: List[str]) -> str:
    """Helper function that will combine a list of links into a regex."""

    return '|'.join(map(re.escape, links))

def get_ids_from_fineweb(
        df_path: str,
        domain: str,
        links: List[str]
) -> List[str]:
    """
    Given a polars DataFrame, filter rows that match one of the target links, and return a list of IDs.

    Args:
        df_path (str): Path to the parquet file where FineWeb data is saved.
        domain (str): Target domain.
        links (List[str]): List of promising links from the target domain.

    Returns:
        List of IDs from FineWeb tahtt align with the target links.
    """
    links = [extract_path(link) for link in links] #Normalize the links.
    regex = combine_links_into_regex(links) #Combine the links into regex for querying FineWeb
    
    #Chooses a column for the initial filter.
    filter_column = 'domain' if 'domain' in pl.read_parquet_schema(df_path) else 'url'

    #loads the FineWeb data with minimal columns
    df = pl.read_parquet(df_path, columns = list(set(['id', 'url', filter_column])))

    #Filters domains by the input domain. Then, filters rows where the URL matches the target link regex.
    filtered = df.filter(df[filter_column].str.contains(domain))
    filtered = filtered.filter(filtered['url'].str.contains(regex, literal=False))

    #Return a list of IDs from the filtered rows.
    return filtered['id'].to_list()

def compare_notes_to_id_files(paths) -> Tuple[dict, List[str]]:
    """Helper function that retrieves notes and ID files and ensures that their content is aligned."""
    
    if not os.path.exists(paths['notes']):
        raise FileNotFoundError("'notes_path' not found. Please input a path to a JSON object with notes for the FineWeb IDs.")
    
    if not str(paths['notes']).endswith('.json'):
        raise ValueError("Input 'notes_path' is not a JSON file.")
    
    if not os.path.exists(paths['ids']):
        raise FileNotFoundError("'ids_dir' not found. Please input a path to a directory with text files containing FineWeb IDs.")
    
    with open(paths['notes'], 'r', encoding='utf-8') as f:
        notes = json.load(f)

    id_files = [file for file in os.listdir(paths['ids']) if file.endswith('.txt')]
    if len(id_files) == 0:
        raise FileNotFoundError("No text files were found in the input 'ids_dir'.")
    
    missing_notes = [note for note in notes.keys() if note not in [file.split('.')[0] for file in id_files]]
    missing_files = [file for file in id_files if file.split('.')[0] not in notes.keys()]

    if missing_notes:
        print(f'{missing_notes} are not recorded in {paths["ids"]}. The final DataFrame may be incomplete.')
    if missing_files:
        print(f'{missing_files} not found in {paths["notes"]}. The final DataFrame may be incomplete.')
    
    if not (missing_files or missing_notes):
        print("Notes and id files match perfectly.")
    return notes, id_files

def id_retrieval_pipeline(
        data_dir: str,
        domain: str,
        links_path: str,
        output_path: str
)-> None:
    """
    Given a domain, a list of links, and a directory where files are saved,
    this pipeline will retrieve IDs from a series of FineWeb files.

    Args:
        data_dir: Path to directory where FineWeb files are saved.
        domain: Target domain for ID extraction.
        links_path: Path to the file where target links are saved.
        output_path: Path to save list of IDs.
    """

    #Load links from the text file.
    links = load_list_from_txt(links_path)

    #Build a list of filepaths.
    paths = [Path(data_dir, file) for file in os.listdir(data_dir)]

    #Iterates through files, extracts IDs, and saves them.
    for path in tqdm(paths, desc = "Finding IDs"):
        ids = get_ids_from_fineweb(path, domain, links)
        save_list_to_txt(ids, output_path)
    
    #Completes the process by returning the number of IDs matched.
    ids = load_list_from_txt(output_path)
    print(f"IDs extracted. {len(ids)} found.")

def build_notes_df(notes: dict) -> pl.DataFrame:
    """Helper function that builds a Polars DataFrame from FineWeb ID notations."""

    flattened = [{'name': key, **subdict} for key, subdict in notes.items()]
    return pl.DataFrame(flattened)

def build_id_df(
        id_dir: str,
        id_files: List[str]
        )-> pl.DataFrame:
    """Helper function that builds a Polars DataFrame from FineWeb IDs."""

    paths = [Path(id_dir, file) for file in id_files]
    id_dict = {path.name.split('.')[0]: load_list_from_txt(path) for path in paths}
    flattened = [{"name": key, "ids": value} for key, value in id_dict.items()]
    output = pl.DataFrame(flattened)

    output = output.with_columns(
        pl.col("ids").map_elements(len, return_dtype=pl.Int32).alias("id_count")
    )
    
    return output

def build_hub_dataset(
        data_dir: str,
        output_path: str='',
        hub_path: str='' 
)-> Union[pl.DataFrame, Dataset]:
    """
    This function combines text files with FineWeb IDs and JSON notes into a DataFrame to be uploaded to HF Hub.

    It is rigidly configured to operate on a data directory (data_dir) with this structure:

    root/
    ├── hub_dataset
            ├── notes.json
            └── ids
                ├──name001.txt
                ├──name002.txt
                └──name003.txt

    'notes.json' is a file with contextual information about the IDs collected from FineWeb.
    Minimally, it must have key values that align with text files in the 'ids' directory.

    'ids' is a direcotry of text files where FineWeb ids are saved.

    Args:
        data_dir (str): Path to directory with data to build HF hub dataset.
        output_path (str): Path where HF hub dataset is to be saved locally.
        hub_path (str): Path to upload HF hub dataset. You must provide an API key for this to function.
    """
    paths = {
        "ids": Path(data_dir, "ids"),
        "notes": Path(data_dir, "notes.json")
    }
    notes, id_files = compare_notes_to_id_files(paths)

    notes = build_notes_df(notes)
    ids = build_id_df(paths['ids'], id_files)
    output = notes.join(ids, on = 'name')

    if output_path:
        output.write_parquet(output_path)
        print(f"Saved to {output_path}")
    
    if hub_path:
        output = Dataset.from_polars(output)
        output.push_to_hub(hub_path)
    return output


from setuptools import setup, find_packages

setup(
    name='fineweb-tools',
    version='0.0.8',
    license='Apache V2',
    description='Tools for preprocessing, analyzing, and distilling FineWeb data.',
    author='Jordan Wolfgang Klein',
    author_email='jwklein14@gmail.com',
    url='https://github.com/user/Lone-Wolfgang',
    keywords=['FineWeb'],
    install_requires=[
        'polars',
        'huggingface-hub',
        'tld',
        'typing',
        'beautifulsoup4',
        'selenium',
        'datasets'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
    packages=find_packages(),
    include_package_data=True
)



from math import log10
import os
from pathlib import Path
import polars as pl
from tqdm.auto import tqdm
from typing import Union, List

def filter_fn(
    path: str,
    filter_column: str,
    target: str,
    retrieve_columns: Union[None, List[str]] = None,
)-> pl.DataFrame:
    """
    Flexible helper function to be used in conjunction with collect_ids.filter_fineweb.

    Given a path to a FineWeb DataFrame, filters by the given parameters.

    Args:
        path(str): Path to a FineWeb DataFrame.
        filter_column(str): Column which is to be filtered.
        target(str): Filters rows where the filter_column contains the target string.
        retrieve_columns(List[str]): Only loads retrieve_columns. If None, retrieves all columns.
    """
    possible_columns = pl.read_parquet_schema(path)
    if not filter_column in possible_columns:
        raise ValueError(f'{filter_column} not found. Please select from {possible_columns}')
    if retrieve_columns and not all(c in possible_columns for c in retrieve_columns):
        raise ValueError(f'Retrieve columns do not align with data. Please select from {possible_columns}')
    df = pl.read_parquet(path, columns=retrieve_columns)
    return df.filter(df[filter_column].str.contains(target))

def filter_fineweb(
    data_dir: str,
    filter_fn: object
)-> pl.DataFrame:
    """
    Iterates through a list of FineWeb files, applies a filter function, and combines results into a single DataFrame.

    Args:
        data(str):Directory where FineWeb parquet files are saved.
        filter_fn(object): Function that takes a filepath, loads it, filters it, and returns a pl.DataFrame.
    """
    output = pl.DataFrame()

    paths = [Path(data_dir, file) for file in os.listdir(data_dir) if file.endswith('.parquet')]
    for path in tqdm(paths, desc='Filtering FineWeb'):
        filtered = filter_fn(path)
        if not isinstance(filtered, pl.DataFrame):
            raise TypeError(f"Input filter function should return a pl.DataFrame. {type(filtered)} returned. Please check your filter function.")
        output = pl.concat([output, filtered])
    return output
    
def group_domains_by_count(
        df: pl.DataFrame,
        group_fn: object = lambda x: int(log10(x))
):
    """
    Grouping function rigidly configured for analysis of the DataFrame yielded by preprocessing.
    Check fineweb_tools.preprocess.download_and_preprocess_pipeline.

    Given a preprocessed DataFrame, groups domains by the number of pages they contributed.

    Args:
        df (pl.DataFrame): Preprocessed dataframe to group.
        group_fn (object): Defaults to grouping by log10.
    """
    # Create the DataFrame and add the 'group' column
    df = df.with_columns(
        pl.col("count").map_elements(group_fn, return_dtype=pl.Int32).alias("group")
    )

    # First, we calculate the total sum of 'count' for the entire dataset
    total_url_sum = df.select(pl.sum("count")).to_numpy()[0][0]

    # Grouping and aggregating
    result = (
        df.group_by("group")
        .agg([
            pl.min("count").alias("group_min"),   # Min of 'count' within the group
            pl.max("count").alias("group_max"),   # Max of 'count' within the group
            pl.count("domain").alias("domains"),  # Count of domain entries per group
            pl.sum("count").alias("pages")   # Sum of 'count' per group
        ])
        .with_columns(
            # Adding the 'corpus_perc' column as the percentage of the total sum
            ((pl.col("pages") / total_url_sum * 100).round(2)).alias("corpus_perc")
        )

    ).sort('group')

    print ('Total Domains:', f"{result['domains'].sum():,}")
    print ('Total URLs:', f"{result['pages'].sum():,}")
    return result